/*
TNKernel real-time kernel - examples

Copyright  2004,2005 Yuri Tiomkin
All rights reserved.

Permission to use, copy, modify, and distribute this software in source
and binary forms and its documentation for any purpose and without fee
is hereby granted, provided that the above copyright notice appear
in all copies and that both that copyright notice and this permission
notice appear in supporting documentation.

THIS SOFTWARE IS PROVIDED BY THE YURI TIOMKIN AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL YURI TIOMKIN OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*============================================================================
  tst1_at91sam.c (demo for AT91SAM7S128)

  Using base TNKernel elements:

    - Tasks
    - Semaphores
    - Data Queues
    - Fixed-Sized Memory Pools

  USART0 is used here with single-buffering DMA for txd and
                           double-buffering DMA for rxd
*===========================================================================*/

#include "AT91SAM7X256.h"
#include "LIB_AT91SAM7X256.H"
#include "utils.h"
#include "tn.h"
#include "tn_port.h"
#include "gui_task.h"
#include "adc_bl_task.h"

//----------- Tasks ----------------------------------------------------------

#define  TASK_USB_PRIORITY		     3
#define  TASK_PROCESSING_PRIORITY    5
#define  TASK_UART_RX_PRIORITY       7
#define  TASK_ADC_BL_PRIORITY       8
#define  TASK_GUI_PRIORITY           10

#define  TASK_PROCESSING_STK_SIZE  128

//-- Big size of stack here case this tasks uses sprintf()
//-- sprintf() has a good appetite for stack(especially in Multi2000 3.5)

#define  TASK_ADC_BL_STK_SIZE     128 //512 //128
#define  TASK_GUI_STK_SIZE     512 //512 //128


unsigned int task_gui_stack[TASK_GUI_STK_SIZE];
unsigned int task_adc_bl_stack[TASK_ADC_BL_STK_SIZE];



TN_TCB  task_adc_bl;
TN_TCB  GUI;


//----------------------------------------------------------------------------
int main()
{
   tn_arm_disable_interrupts();

   HardwareInit();

   tn_start_system(); //-- Never returns

   return 1;
}

//----------------------------------------------------------------------------
void  tn_app_init()
{
   //--- Task UART TX

   task_adc_bl.id_task = 0;
   tn_task_create(&task_adc_bl,                 //-- task TCB
                 task_adc_bl_func,              //-- task function
                 TASK_ADC_BL_PRIORITY,          //-- task priority
                 &(task_adc_bl_stack            //-- task stack first addr in memory
                    [TASK_ADC_BL_STK_SIZE-1]),
                 TASK_ADC_BL_STK_SIZE,          //-- task stack size (in int,not bytes)
                 NULL,                           //-- task function parameter
                 TN_TASK_START_ON_CREATION       //-- Creation option
                 );

   //--- Task IO
   GUI.id_task = 0;
   tn_task_create(&GUI,                      //-- task TCB
                 task_gui_func,                   //-- task function
                 TASK_GUI_PRIORITY,               //-- task priority
                 &(task_gui_stack                 //-- task stack first addr in memory
                    [TASK_GUI_STK_SIZE-1]),
                 TASK_GUI_STK_SIZE,               //-- task stack size (in int,not bytes)
                 NULL,                           //-- task function parameter
                 TN_TASK_START_ON_CREATION       //-- Creation option
                 );	   

}




//----------------------------------------------------------------------------



