/*
TNKernel real-time kernel - examples

Copyright  2004,2005 Yuri Tiomkin
All rights reserved.

Permission to use, copy, modify, and distribute this software in source
and binary forms and its documentation for any purpose and without fee
is hereby granted, provided that the above copyright notice appear
in all copies and that both that copyright notice and this permission
notice appear in supporting documentation.

THIS SOFTWARE IS PROVIDED BY THE YURI TIOMKIN AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL YURI TIOMKIN OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

#include "AT91SAM7X256.h"
#include "LIB_AT91SAM7X256.H"
#include "utils.h"
#include "tn.h"

/*-------- AT91SAM7S128 - project's specific pinout description ---------------

                                        I/O dir set            I/O dir
                                       -------------          ---------

 PA0   - LED7                              manual                out
 PA1   - not uses
 PA2   - LED8
 PA3   - SDA
 PA4   - SCL
 PA5   - RXD0
 PA6   - TXD0
 PA7   - not uses
 PA8   - not uses
 PA9   - not uses
 PA10  - LED4
 PA11  - NPCS0 (SPI CS)
 PA12  - MISO
 PA13  - MOSI
 PA14  - SCK
 PA15  - LED2
 PA16  - LED1
 PA17  - button
 PA18  - not uses
 PA19  - not uses
 PA20  - not uses
 PA21  - RXD1
 PA22  - TXD1
 PA23  - not uses
 PA24  - USB-D
 PA25  - DP-UP
 PA26  - LED3
 PA27  - not uses
 PA28  - not uses
 PA29  - LED5
 PA30  - LED6
 PA31  - not uses

----------------------------------------------------------------------------*/

void tn_cpu_irq_handler(void);
void spurious_int_func(void);

//----------------------------------------------------------------------------
void HardwareInit(void)
{
AT91PS_AIC pAic=AT91C_BASE_AIC;
AT91PS_PWMC      pPWM=AT91C_BASE_PWMC;;
  //-- Enable pin control


// First, enable the clock of the PIO
       AT91F_PMC_EnablePeriphClock ( AT91C_BASE_PMC, (1 << AT91C_ID_PIOA)|(1 << AT91C_ID_PIOB) ) ;
       AT91F_PIO_CfgOutput( AT91C_BASE_PIOA, AT91C_PIO_PA15| 	//pa15 lcd_cs
	   										 AT91C_PIO_PA14| 	//lcd_reset
	   										 AT91C_PIO_PA13);	//pa14 lcd_RS

//configure ADC
AT91F_PMC_EnablePeriphClock(AT91C_BASE_PMC,(1 << AT91C_ID_ADC)); // PIO controller base address		
AT91F_ADC_SoftReset(AT91C_BASE_ADC);
AT91F_ADC_CfgModeReg(AT91C_BASE_ADC,(5<<8)|
									(7<<16)|
									10<<24);
//Configure PWM
AT91F_PMC_EnablePeriphClock(AT91C_BASE_PMC,(1 << AT91C_ID_PWMC));//enable clock
AT91F_PIO_CfgPeriph(AT91C_BASE_PIOB,0,AT91C_PB27_PWM0);//selec periph


	pPWM->PWMC_CH[0].PWMC_CDTYR = 10;
	pPWM->PWMC_CH[0].PWMC_CPRDR = 255;
	pPWM->PWMC_CH[0].PWMC_CMR = 1<<9;
	pPWM->PWMC_ENA = 1;


  //-- Setup  Periodic Interval Timer for interrupt 1 ms  55MHz
  AT91F_PITInit(AT91C_BASE_PITC,1001,55);//-- PIT Enable,Periodic Interval Value - 1 ms
  AT91F_PITEnableInt(AT91C_BASE_PITC);//-- PIT Interrupt Enable


  // Setup System Interrupt Mode and Vector with Priority 7 and Enable it */
  // PIT's int is System Interrupt,Peripheral ID for PIT is 1
  AT91F_AIC_CfgPMC();	//Enable Peripheral clock in PMC for AIC

 	AT91F_AIC_ConfigureIt ( AT91C_BASE_AIC, AT91C_ID_SYS, 7,AT91C_AIC_SRCTYPE_INT_POSITIVE_EDGE, tn_cpu_irq_handler);
	AT91F_AIC_EnableIt (AT91C_BASE_AIC, AT91C_ID_SYS);



  //AT91PS_AIC->AIC_SMR1 = (1 << 5) | 7;         // AT91C_ID_SYS - System Peripheral
 // AIC_IECR = (1<<1);               //-- Enable System Peripheral Int



  pAic->AIC_SPU  = (unsigned int) spurious_int_func;

}

//----------------------------------------------------------------------------


