;------------------------------------------------------------------------------
;
;    TNKernel startup hardware init for AT91SAM7S128 processor
;    (for ARM(c) compiler)
;
;    ARM(c) assembler
;
; Copyright  2004,2005 Yuri Tiomkin
; All rights reserved.
;
;Permission to use, copy, modify, and distribute this software in source
;and binary forms and its documentation for any purpose and without fee
;is hereby granted, provided that the above copyright notice appear
;in all copies and that both that copyright notice and this permission
;notice appear in supporting documentation.
;
;THIS SOFTWARE IS PROVIDED BY THE YURI TIOMKIN AND CONTRIBUTORS ``AS IS'' AND
;ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;ARE DISCLAIMED. IN NO EVENT SHALL YURI TIOMKIN OR CONTRIBUTORS BE LIABLE
;FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;SUCH DAMAGE.

;  ver 2.0


  ;-- Public functions declared in this file

     EXPORT tn_startup_hardware_init

  ;-- Extern reference

     IMPORT reset

   ;-- Constants


MC_FMR    EQU  0xFFFFFF60
WDT_MR    EQU  0xFFFFFD44
CKGR_MOR  EQU  0xFFFFFC20
PMC_SR    EQU  0xFFFFFC68
CKGR_PLLR EQU  0xFFFFFC2C
PMC_MCKR  EQU  0xFFFFFC30
MC_RCR    EQU  0xFFFFFF00

NOINT     EQU  0xc0

;----------------------------------------------------------------------------

           AREA    HardwareInit, CODE, READONLY

           CODE32

;----------------------------------------------------------------------------
;  This routine is called immediately after reset to setup hardware that is
; vital for processor's functionality (for instance,SDRAM controller setup,
; PLL setup,etc.)
;  It is assumed that other hardware's init routine(s) will be invoked later
; by C-language function call.
;----------------------------------------------------------------------------

tn_startup_hardware_init

    ;-- Set flash waite state
    ;-- bits 23..16 - the number of Master Clock cycles in 1
    ;-- microsecond. For f = 48 MHz this field's value setted to 48
    ;-- bits 9..8  01 - rd = 2 cycles,wr = 3 cycles
    ;-- rMC_FMR = (48 << 16) | (1 << 8);

     ldr  r0, =(48 << 16) | (1 << 8)
     ldr  r3, =MC_FMR
     str  r0, [r3]

    ;-- Disable watchdog
    ;-- rWDT_MR = (1<<15);    ;-- bit 15 = 1 - disable WDT

     ldr  r0, =(1<<15)
     ldr  r3, =WDT_MR
     str  r0, [r3]

    ;-- Set main clock at 48000000 MHz ( for crystal 16 MHz)
    ;-- Main osc start-up time according  manual's table 35-8 will be
    ;-- max 1.5 mS. => bits 15..8 - (1500 /(30.51)) / 8 ~= 7
    ;-- rCKGR_MOR =  (7 << 8)  | 1;  ;-- bit 0=1 The Main Oscillator is enabled.

     ldr  r0, =(7 << 8) | 1
     ldr  r3, =CKGR_MOR
     str  r0, [r3]

    ;-- Wait while main oscillator is not stabilized.
    ;-- bit 0 of rPMC_SR =1 - Main oscillator is stabilized.

l_1
     ldr  r0, =PMC_SR
     ldr  r1,[r0]
     tst  r1, #0x00000001
     beq  l_1


    ;-- DIV - by 2 (bypassed);   Fin = 16/2 = 8 MHz
    ;-- MUL - (11+1); Fout =  8*(11+1) = 96 MHz
    ;-- PLLCOUNT  - 28     (conservative value)
    ;-- rCKGR_PLLR =    2 |         ;-- bits 7..0   (DIV) = 2;
    ;         (11<<16) |   ;-- bits 26..16 (MUL) = 11;
    ;         (28 << 8);   ;-- bits 13..8 (PLLCOUNT)   = 28

     ldr  r0, =(11<<16) | (28 << 8) | 2
     ldr  r3, =CKGR_PLLR
     str  r0, [r3]

    ;-- Wait while PLL is not locked (bit 2 of rPMC_SR =1 - PLL is locked)

l_2
     ldr  r0, =PMC_SR
     ldr  r1,[r0]
     tst  r1, #0x00000004
     beq  l_2


  ;-- Select PLL as clock source - Master clock is PLL clock divided by 2
  ;--  rPMC_MCKR = (0x3) |(1<<2) ; ;-- bits 4..2 = 001 - Selected clock divided by 2
  ;                                ;-- bits 1..0 = 11 - PLL Clock is selected as Master clock

     ldr  r0, =(1<<2) | 0x03
     ldr  r3, =PMC_MCKR
     str  r0, [r3]

  ;-- Wait while Master Clock is not ready (bit 3 of rPMC_SR =1 - Master Clock is ready)

l_3
     ldr  r0, =PMC_SR
     ldr  r1,[r0]
     tst  r1, #0x00000008
     beq  l_3
     bx   lr

;----------------------------------------------------------------------------
      EXPORT  tn_arm_disable_interrupts

tn_arm_disable_interrupts

     mrs  r0, cpsr
     orr  r0, r0, #NOINT
     msr  cpsr_c, r0
     bx   lr


     EXPORT  tn_arm_enable_interrupts


tn_arm_enable_interrupts

     mrs  r0, cpsr
     bic  r0, r0, #NOINT
     msr  cpsr_c, r0
     bx   lr


     END

;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------




