;------------------------------------------------------------------------------
; TNKernel real-time kernel
;
; Copyright  2004,2007 Yuri Tiomkin
; All rights reserved.
;
; ffs_asm() - this is the ffs algorithm devised by D.Seal and posted to
;             comp.sys.arm on  16 Feb 1994.
; Interrupt context switch -  this source code is derived on code
;             written by WellsK
;
; Permission to use, copy, modify, and distribute this software in source
; and binary forms and its documentation for any purpose and without fee
; is hereby granted, provided that the above copyright notice appear
; in all copies and that both that copyright notice and this permission
; notice appear in supporting documentation.
;
; THIS SOFTWARE IS PROVIDED BY THE YURI TIOMKIN AND CONTRIBUTORS "AS IS" AND
; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL YURI TIOMKIN OR CONTRIBUTORS BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
; SUCH DAMAGE.
;------------------------------------------------------------------------------

;--- ver 2.4 ---

   ;---- ARM(c) assembler  ----

        PRESERVE8

  ;-- External references

  IMPORT  tn_curr_run_task
  IMPORT  tn_next_task_to_run
  IMPORT  tn_int_counter
  IMPORT  tn_context_switch_request
;  IMPORT  tn_cpu_irq_handler
  IMPORT  tn_system_state


  ;-- Public functions declared in this file

  EXPORT  tn_switch_context_exit
  EXPORT  tn_switch_context
  EXPORT  tn_cpu_irq_isr
;  EXPORT  tn_cpu_fiq_isr
  EXPORT  tn_cpu_save_sr
  EXPORT  tn_cpu_restore_sr
  EXPORT  tn_start_exe
  EXPORT  tn_chk_irq_disabled
  EXPORT  OSPendSV
;  EXPORT  ffs_asm

  ;-- Constants

NVIC_INT_CTRL   EQU     0xE000ED04               ; interrupt control state register
NVIC_SYSPRI2    EQU     0xE000ED20               ; system priority register (2)
NVIC_PENDSV_PRI EQU     0x00000000               ; PendSV priority value (lowest)
NVIC_PENDSVSET  EQU     0x10000000               ; value to trigger PendSV exception


;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
        AREA |.text|, CODE, READONLY, ALIGN=2
		THUMB
		REQUIRE8
		PRESERVE8

;----------------------------------------------------------------------------
; Interrups not yet enabled
;----------------------------------------------------------------------------
tn_start_exe

        LDR     R4, =NVIC_SYSPRI2       ; set the PendSV exception priority
        LDR     R5, =NVIC_PENDSV_PRI
        STR     R5, [R4]

        MOV     R4, #0                  ; set the PSP to 0 for initial context switch call
        MSR     PSP, R4

        LDR     R4, =tn_system_state        ; OSRunning = TRUE
        MOV     R5, #1
        STRB    R5, [R4]

        LDR     R4, =NVIC_INT_CTRL      ; trigger the PendSV exception (causes context switch)
        LDR     R5, =NVIC_PENDSVSET
        STR     R5, [R4]

        CPSIE   I                       ; enable interrupts at processor level
OSStartHang
        B       OSStartHang             ; should never get here

;     ldr    r1,=tn_system_state    ; Indicate that system has started
;     mov    r0,#1                  ; 1 -> TN_SYS_STATE_RUNNING
;     strb   r0,[r1]

tn_switch_context_exit

 ;    ldr    r0, =tn_curr_run_task
 ;    ldr    r1, =tn_next_task_to_run
;     ldr    r1, [r1]                   ; get stack pointer
;     ldr    sp, [r1]                   ; switch to the new stack
;     str    r1, [r0]                   ; set new current running task tcb address

;     ldmfd  sp!, {r0}
;     msr    SPSR_cxsf, r0
;     ldmfd  sp!, {r0-r12,lr,pc}^

;----------------------------------------------------------------------------
tn_switch_context

        LDR     R4, =NVIC_INT_CTRL      ; trigger the PendSV exception (causes context switch)
        LDR     R5, =NVIC_PENDSVSET
        STR     R5, [R4]
        BX      LR
;
;----------------------------------------------------------------------------
OSPendSV

        MRS     R0, PSP                 ; PSP is process stack pointer
        CBZ     R0, OSPendSV_nosave     ; skip register save the first time

        SUB     R0, R0, #0x20           ; save remaining regs r4-11 on process stack
        STM     R0, {R4-R11}

        LDR     R4, =tn_curr_run_task         ; OSTCBCur->OSTCBStkPtr = SP;
        LDR     R4, [R4]
        STR     R0, [R4]                ; R0 is SP of process being switched out



;     stmfd  sp!, {lr}                  ; Save return address
;     stmfd  sp!, {r0-r12,lr}           ; Save curr task registers
;     mrs    r0,  cpsr
;     tst    LR, #1                     ; from Thumb mode ?
;     orrne  R0,  R0, #0x20             ; set the THUMB bit
;     stmfd  sp!, {r0}                  ; Save current CPSR

;     mrs    r0,  cpsr
;     orr    r0, r0, #NOINT             ; Disable Int
;     msr    CPSR_c, r0
OSPendSV_nosave
        LDR     R4, =tn_curr_run_task         ; OSTCBCur  = OSTCBHighRdy;
        LDR     R6, =tn_next_task_to_run
        LDR     R6, [R6]
        STR     R6, [R4]

        LDR     R0, [R6]                ; R0 is new process SP; SP = OSTCBHighRdy->OSTCBStkPtr;
        LDM     R0, {R4-R11}            ; restore r4-11 from new process stack
        ADD     R0, R0, #0x20
        MSR     PSP, R0                 ; load PSP with new process SP
        ORR     LR, LR, #0x04           ; ensure exception return uses process stack
        BX      LR                      ; exception return will restore remaining context

        NOP
;     ldr    r1,=tn_curr_run_task
;    ldr    r2, [r1]
;     ldr    r0,=tn_next_task_to_run
;     ldr    r0, [r0]
;     cmp    r0,#0                ; sanity checking
;     beq    tn_sw_restore
;     cmp    r2,r0
;     beq    tn_sw_restore

;     str    sp, [r2]            ; store sp in preempted tasks's TCB
;     ldr    sp, [r0]            ; get new task's sp
;     str    r0, [r1]            ; tn_curr_run_task = tn_next_task_to_run

tn_sw_restore

;     ldmfd   sp!, {r0}
;     msr     SPSR_cxsf, r0
;     ldmfd   sp!, {r0-r12,lr,pc}^

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
tn_cpu_irq_isr

 ;    sub    lr, lr, #4              ; Set lr to the actual return address
 ;    stmfd  sp!, {r0-r12, lr}       ; save all registers
 ;
 ;    ldr    r2, =tn_int_counter
 ;    ldr    r3, [r2]                ; interrupt_counter -> r3
 ;    add    r0, r3,#1               ; interrupt_counter++
 ;    str    r0, [r2]

 ;    LDR     R0,=tn_cpu_irq_handler
 ;    MOV     LR, PC
  ;   BX      R0

     ; interrupt_counter--

;     ldr    r2, =tn_int_counter
;     ldr    r1, [r2]
;     sub    r3, r1,#1
;     str    r3, [r2]
;     cmp    r3, #0x00               ; if it is nested int - return
;     bne    exit_irq_int            ; return from int
;     ldr    r0, =tn_context_switch_request  ; see if we need to do a context switch
;     ldr    r1, [R0]
;     cmp    r1, #0                  ; if 0 - return
;     beq    exit_irq_int
;     mov    r1, #0                  ; else - clear req flag and  goto context switch
;     str    r1, [r0]
;     b      tn_int_ctx_switch

exit_irq_int

;     ldmfd  sp!, {r0-r12, pc}^      ; exit

;---------------------------------------------------------------------------

tn_int_ctx_switch

        LDR     R4, =NVIC_INT_CTRL      ; trigger the PendSV exception (causes context switch)
        LDR     R5, =NVIC_PENDSVSET
        STR     R5, [R4]
        BX      LR

        NOP
  ; Our target - get all registers of interrrupted task, saved in IRQ's stack
  ; and save them in the interrupted task's stack


 ;    mrs    r0, spsr                  ; Get interrupted task's CPRS
 ;    stmfd  sp!, {r0}                 ; Save it in the IRQ stack
 ;    add    sp, sp, #4                ; Restore stack ptr after CPSR saving
 ;    ldmfd  sp!, {r0-r12, lr}         ; Put all saved registers from IRQ
                                      ;   stack back to registers
 ;    mov    r1, sp                    ; r1 <-  IRQ's SP
 ;    sub    r1,r1,#4
 ;    msr    cpsr_c, #(NOINT | SVCMODE) ; Change to SVC mode; INT are disabled

   ; Now - in SVC mode; in r1 - IRQ's SP

 ;    ldr    r0, [r1], #-14*4    ; r0 <- task's lr (instead pc)+ rewind stack ptr
 ;    stmfd  sp!, {r0}
 ;    stmfd  sp!, {r2-r12, lr}   ; Save registers in interrupted task's
                                ; stack,except CPSR,r0,r1
 ;    ldr    r0, [r1],#2*4       ; Get interrupted task's CPSR (pos 0(-15))
 ;    ldr    r2, [r1],#-4        ; Get valid r1 to save (pos 2(-13))
 ;    ldr    r1, [r1]            ; Get valid r0 to save (pos 1(-14))
 ;    stmfd  sp!, {r0-r2}        ; Save r0, r1, and CPSR

  ; Registers has been saved. Now - switch context

 ;    ldr    r0, =tn_curr_run_task
 ;    ldr    r0, [r0]
 ;    str    sp, [r0]               ; SP <- curr task

 ;    ldr    r0, =tn_next_task_to_run
 ;    ldr    r2, [r0]
 ;    ldr    r0, =tn_curr_run_task
 ;    str    r2, [r0]
 ;    ldr    sp, [r2]

  ; Return

  ;   ldmfd  sp!, {r0}                  ; Get CPSR
  ;   msr    spsr_cxsf, r0              ; SPSR <- CPSR
  ;   ldmfd  sp!, {r0-r12, lr, pc}^     ; Restore all registers, CPSR also

;----------------------------------------------------------------------------
;    Do nothing here
;----------------------------------------------------------------------------
;tn_cpu_fiq_isr

;     stmfd  SP!,{R0-R12,LR}
;     ldmfd  SP!,{R0-R12,LR}      ; restore registers of interrupted task's stack
;     subs   PC,LR,#4             ; return from FIQ

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
tn_cpu_save_sr

        MRS     R0, PRIMASK                 ; set prio int mask to mask all (except faults)
        CPSID   I
        BX      LR

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
tn_cpu_restore_sr

        MSR     PRIMASK, R0
        BX      LR

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
tn_chk_irq_disabled

     MRS     R0, PRIMASK
     and    r0,r0,#1
     bx     lr

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
;ffs_asm

      ; Standard trick to isolate bottom bit in r0 or 0 if r0 = 0 on entry
;        rsb    r1, r0, #0
;        ands   r0, r0, r1


         ; now r0 has at most one set bit, call this X
         ; if X = 0, all further instructions are skipped

;        adrne  r2, L_ffs_table
;        orrne  r0, r0, r0, lsl #4       ; r0 = X * 0x11
 ;       orrne  r0, r0, r0, lsl #6       ; r0 = X * 0x451
 ;       rsbne  r0, r0, r0, lsl #16      ; r0 = X * 0x0450fbaf

        ; now lookup in table indexed on top 6 bits of r0
;        ldrneb r0, [ r2, r0, lsr #26 ]
 ;       bx     lr

;L_ffs_table
              ;    0   1   2   3   4   5   6   7
;        DCB        0,  1,  2, 13,  3,  7,  0, 14  ;  0- 7
;        DCB        4,  0,  8,  0,  0,  0,  0, 15  ;  8-15
;        DCB       11,  5,  0,  0,  9,  0,  0, 26  ; 16-23
;        DCB        0,  0,  0,  0,  0, 22, 28, 16  ; 24-31
;        DCB       32, 12,  6,  0,  0,  0,  0,  0  ; 32-39
;        DCB       10,  0,  0, 25,  0,  0, 21, 27  ; 40-47
;        DCB       31,  0,  0,  0,  0, 24,  0, 20  ; 48-55
;        DCB       30,  0, 23, 19, 29, 18, 17,  0  ; 56-63
;----------------------------------------------------------------------------
      EXPORT  tn_arm_disable_interrupts

tn_arm_disable_interrupts

        CPSID   I
        BX      LR


     EXPORT  tn_arm_enable_interrupts


tn_arm_enable_interrupts

        CPSIE   I
        BX      LR


     END

;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------




